/***************************************************************************
 *   Copyright (C) 2007  Antony Martin <martin.antony@yahoo.fr>            *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or         *
 *   modify it under the terms of the GNU General Public License           *
 *   as published by the Free Software Foundation; either version 2        *
 *   of the License, or (at your option) any later version.                *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA          *
 ***************************************************************************/

/* Cree le : 16 septembre 2006
   derniere modification le 18/03/2007 */


#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>

#include <SCE/utils/SCEString.h>
#include <SCE/utils/SCETime.h>

#include <SCE/utils/SCEError.h>

/* taille maximale des chaines de caractere */
#define SCE_MAXLEN 4096

typedef struct sce_errorsrc SCE_ErrorSrc;
struct sce_errorsrc
{
    unsigned int line;
    char func[SCE_MAXLEN];
    char file[SCE_MAXLEN];
};

/* structure erreur: contient des informations sur une erreur */
typedef struct sce_error SCE_Error;
struct sce_error
{
    time_t date;
    
    int have_error;
    int have_msg;
    int code;
    
    unsigned int line;
    char func[SCE_MAXLEN];
    char file[SCE_MAXLEN];
    
    char msg[SCE_MAXLEN];
    
    size_t n_src;
    SCE_ErrorSrc *src;
};

static SCE_Error error;
static FILE* stream;


int SCEError_Init(FILE *outlog)
{
    unsigned int i;
    
    /* initialisation des tampons */
    for(i=0; i<SCE_MAXLEN; i++)
    {
        error.msg[i] = '\0';
        error.file[i] = '\0';
        error.func[i] = '\0';
    }
    
    error.date = 0;
    error.have_error = 0;
    error.have_msg = 0;
    error.code = 0;
    error.line = 0;
    error.n_src = 0;
    if(error.src)
        free(error.src);
    error.src = NULL;
    
    stream = outlog;
    
    if(stream == NULL)
        return (-1);
    
    return (0);
}

void SCEError_Clear(void)
{
    if(error.src)
        free(error.src);
    
    SCEError_Init(stream);
}

/* stock un code d'erreur */
void
SCEError_SetErrorCode_(
  const char *file,     /* fichier ou s'est produite l'erreur */
  const char *func,     /* fonction */
  const int line,       /* ligne du fichier */
  int code)       /* code d'erreur */
{
    error.date = time(NULL);
    error.have_error = 1;
    error.line = line;
    error.code = code;
    
    /* pas de verification de la taille de file et func, en principe
       aucun nom de fichier/fonction ne depasse SCE_MAXLEN caracteres... */
    strcpy(error.file, file);
    strcpy(error.func, func);
}


int
Logger_LogMsg(const char *fmt, ...)
{
    va_list args;
    
    if(error.have_msg)
        return (error.code);
    
    if(!error.have_error)
        error.date = time(NULL);
    
    error.have_error = 1;
    error.have_msg = 1;
    
    /* NOTE: pour bien faire, il faudrait verifier la taille
             de fmt et de ses arguments */
    va_start(args, fmt);
    vsprintf(error.msg, fmt, args);
    va_end(args);
    
    return (0);
}

void
Logger_PrintMsg(const char *fmt, ...)
{
    va_list args;
    
    va_start(args, fmt);
    vfprintf(stream, fmt, args);
    va_end(args);
}

void
SCEError_SetSrc(const char *file, const char *func, const int line)
{
    error.n_src++;
    error.src = realloc(error.src, error.n_src*sizeof *error.src);
    if(error.src == NULL)
    {
        fputs("error: cannot allocate memory in error manager!\n"
              "       exit forced.\n", stderr);
        exit(EXIT_FAILURE);
    }
    error.src[error.n_src-1].line = line;
    strcpy(error.src[error.n_src-1].file, file);
    strcpy(error.src[error.n_src-1].func, func);
}



const char* SCEError_GetErrorMsg(void)
{
    return (error.msg);
}


int SCEError_GetErrorCode(void)
{
    return (error.code);
}


const char* SCEError_GetCodeMsg(int code)
{
    switch(code)
    {
        case SCE_OUT_OF_MEMORY:
            strcpy(error.msg, "out of memory");
            break;
        
        case SCE_INVALID_ARG:
            strcpy(error.msg, "invalid argument");
            break;
        
        case SCE_INVALID_OPERATION:
            strcpy(error.msg, "invalid operation");
            break;
        
        case SCE_INVALID_POINTER:
            strcpy(error.msg, "invalid pointer");
            break;
        
        case SCE_INVALID_SIZE:
            strcpy(error.msg, "invalid size");
            break;
        
        case SCE_FILE_NOT_FOUND:
            strcpy(error.msg, "file not found");
            break;
        
        case SCE_BAD_FORMAT:
            strcpy(error.msg, "bad format");
            break;
        
        default:
            strcpy(error.msg, "unknown error");
            break;
    }
    
    return (error.msg);
}


const char* SCEError_GetFile(void)
{
    return (error.file);
}

const char* SCEError_GetFunc(void)
{
    return (error.func);
}


int SCEError_GetLine(void)
{
    return (error.line);
}


time_t SCEError_GetTime(void)
{
    return (error.date);
}


int SCEError_HaveError(void)
{
    return (error.have_error);
}


int SCEError_HaveMsg(void)
{
    return (error.have_msg);
}


int SCEError_OutStream(void)
{
    char *indent = NULL;
    
    /* important: posind doit etre initialiser a zero */
    unsigned int posind = 0;
    int i = 0;
    /* structure des donnees de la date */
    const struct tm const *time_info = NULL;
    char date[25];
    
    if(stream == NULL)
    {
        fputs("error: invalid output stream!\n", stderr);
        return (-1);
    }
    
    indent = malloc(((error.n_src+2)*2)+1);
    if(indent == NULL)
    {
        fputs("error: cannot allocate memory in error manager!\n", stderr);
        return (-1);
    }
    
    /* initialisation des chaines de caractere */
    memset(indent, '\0', ((error.n_src+2)*2)+1);
    memset(date, '\0', sizeof date);
    
    time_info = gmtime(&error.date);
    SCE_Time_MakeString(date, time_info);
    
    fprintf(stream, "\n[log %s ]\n>>\n",date);
    
    /* macro "d'indentation", incremente la chaine d'indentation */
    #define INDENT()\
        indent[posind] = indent[posind+1] = ' ';\
        posind += 2;
    
    for(i=error.n_src-1; i>=0; i--)
    {
        INDENT()
        
        fprintf(stream,
        "%sfile: '%s'\n"
        "%sfunction: '%s'\n"
        "%sline: %d\n"
        "%s>>\n",
        indent, error.src[i].file,
        indent, error.src[i].func,
        indent, error.src[i].line,
        indent);
    }
    
    INDENT()
    
    fprintf(stream,
    "%sfile: '%s'\n"
    "%sfunction: '%s'\n"
    "%sline: %d\n"
    "%s>>\n",
    indent, error.file,
    indent, error.func,
    indent, error.line,
    indent);
    
    INDENT()
    
    #undef INDENT
    
    fprintf(stream,
    "%serror:\n"
    "%scode: %d\n"
    "%smessage: %s\n"
    "[end]\n",
    indent,
    indent, error.code,
    indent, error.msg);
    
    free(indent);
    indent = NULL;
    
    return (0);
}
